# Copyright (c) 2012 OpenStack Foundation.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import netaddr
import webob.exc

from oslo_log import log as logging
from oslo_utils import strutils

from tacker._i18n import _
from tacker.api import api_common
from tacker.api.v1 import attributes
from tacker.api.v1 import resource as wsgi_resource
from tacker.common import exceptions
from tacker.common import rpc as n_rpc
from tacker import policy


LOG = logging.getLogger(__name__)

FAULT_MAP = {exceptions.NotFound: webob.exc.HTTPNotFound,
             exceptions.Conflict: webob.exc.HTTPConflict,
             exceptions.InUse: webob.exc.HTTPConflict,
             exceptions.BadRequest: webob.exc.HTTPBadRequest,
             exceptions.ServiceUnavailable: webob.exc.HTTPServiceUnavailable,
             exceptions.NotAuthorized: webob.exc.HTTPForbidden,
             netaddr.AddrFormatError: webob.exc.HTTPBadRequest,
             }


class Controller(object):
    LIST = 'list'
    SHOW = 'show'
    CREATE = 'create'
    UPDATE = 'update'
    DELETE = 'delete'

    def __init__(self, plugin, collection, resource, attr_info,
                 allow_bulk=False, member_actions=None, parent=None,
                 allow_pagination=False, allow_sorting=False):
        member_actions = member_actions or []
        self._plugin = plugin
        self._collection = collection.replace('-', '_')
        self._resource = resource.replace('-', '_')
        self._attr_info = attr_info
        self._allow_bulk = allow_bulk
        self._allow_pagination = allow_pagination
        self._allow_sorting = allow_sorting
        self._native_bulk = self._is_native_bulk_supported()
        self._native_pagination = self._is_native_pagination_supported()
        self._native_sorting = self._is_native_sorting_supported()
        self._policy_attrs = [name for (name, info) in self._attr_info.items()
                              if info.get('required_by_policy')]
        self._notifier = n_rpc.get_notifier('nfv')
        self._member_actions = member_actions
        self._primary_key = self._get_primary_key()
        if self._allow_pagination and self._native_pagination:
            # Native pagination need native sorting support
            if not self._native_sorting:
                raise exceptions.Invalid(
                    _("Native pagination depend on native sorting")
                )
            if not self._allow_sorting:
                LOG.info("Allow sorting is enabled because native "
                         "pagination requires native sorting")
                self._allow_sorting = True

        if parent:
            self._parent_id_name = '%s_id' % parent['member_name']
            parent_part = '_%s' % parent['member_name']
        else:
            self._parent_id_name = None
            parent_part = ''
        self._plugin_handlers = {
            self.LIST: 'get%s_%s' % (parent_part, self._collection),
            self.SHOW: 'get%s_%s' % (parent_part, self._resource)
        }
        for action in [self.CREATE, self.UPDATE, self.DELETE]:
            self._plugin_handlers[action] = '%s%s_%s' % (action, parent_part,
                                                         self._resource)

    def _get_primary_key(self, default_primary_key='id'):
        for key, value in (self._attr_info).items():
            if value.get('primary_key', False):
                return key
        return default_primary_key

    def _is_native_bulk_supported(self):
        native_bulk_attr_name = ("_%s__native_bulk_support"
                                 % self._plugin.__class__.__name__)
        return getattr(self._plugin, native_bulk_attr_name, False)

    def _is_native_pagination_supported(self):
        native_pagination_attr_name = ("_%s__native_pagination_support"
                                       % self._plugin.__class__.__name__)
        return getattr(self._plugin, native_pagination_attr_name, False)

    def _is_native_sorting_supported(self):
        native_sorting_attr_name = ("_%s__native_sorting_support"
                                    % self._plugin.__class__.__name__)
        return getattr(self._plugin, native_sorting_attr_name, False)

    def _exclude_attributes_by_policy(self, context, data):
        """Identifies attributes to exclude according to authZ policies.

        Return a list of attribute names which should be stripped from the
        response returned to the user because the user is not authorized
        to see them.
        """
        attributes_to_exclude = []
        for attr_name in data:
            attr_data = self._attr_info.get(attr_name)
            if attr_data and attr_data['is_visible']:
                if policy.check(
                    context,
                    '%s:%s' % (self._plugin_handlers[self.SHOW], attr_name),
                    data,
                        might_not_exist=True):
                    # this attribute is visible, check next one
                    continue
            # if the code reaches this point then either the policy check
            # failed or the attribute was not visible in the first place
            attributes_to_exclude.append(attr_name)
        return attributes_to_exclude

    def _view(self, context, data, fields_to_strip=None):
        """Build a view of an API resource.

        :param context: the tacker context
        :param data: the object for which a view is being created
        :param fields_to_strip: attributes to remove from the view

        :returns: a view of the object which includes only attributes
        visible according to API resource declaration and authZ policies.
        """
        fields_to_strip = ((fields_to_strip or []) +
                           self._exclude_attributes_by_policy(context, data))
        return self._filter_attributes(context, data, fields_to_strip)

    def _filter_attributes(self, context, data, fields_to_strip=None):
        if not fields_to_strip:
            return data
        return dict(item for item in (data).items()
                    if (item[0] not in fields_to_strip))

    def _do_field_list(self, original_fields):
        fields_to_add = None
        # don't do anything if fields were not specified in the request
        if original_fields:
            fields_to_add = [attr for attr in self._policy_attrs
                             if attr not in original_fields]
            original_fields.extend(self._policy_attrs)
        return original_fields, fields_to_add

    def __getattr__(self, name):
        if name in self._member_actions:
            def _handle_action(request, id, **kwargs):
                arg_list = [request.context, id]
                # Ensure policy engine is initialized
                policy.init()
                # Fetch the resource and verify if the user can access it
                try:
                    resource = self._item(request, id, True)
                except exceptions.PolicyNotAuthorized:
                    msg = _('The resource could not be found.')
                    raise webob.exc.HTTPNotFound(msg)
                body = kwargs.pop('body', None)
                # Explicit comparison with None to distinguish from {}
                if body is not None:
                    arg_list.append(body)
                # It is ok to raise a 403 because accessibility to the
                # object was checked earlier in this method
                policy.enforce(request.context, name, resource)
                return getattr(self._plugin, name)(*arg_list, **kwargs)
            return _handle_action
        else:
            raise AttributeError

    def _get_pagination_helper(self, request):
        if self._allow_pagination and self._native_pagination:
            return api_common.PaginationNativeHelper(request,
                                                     self._primary_key)
        elif self._allow_pagination:
            return api_common.PaginationEmulatedHelper(request,
                                                       self._primary_key)
        return api_common.NoPaginationHelper(request, self._primary_key)

    def _get_sorting_helper(self, request):
        if self._allow_sorting and self._native_sorting:
            return api_common.SortingNativeHelper(request, self._attr_info)
        elif self._allow_sorting:
            return api_common.SortingEmulatedHelper(request, self._attr_info)
        return api_common.NoSortingHelper(request, self._attr_info)

    def _items(self, request, do_authz=False, parent_id=None):
        """Retrieves and formats a list of elements of the requested entity."""
        # NOTE(salvatore-orlando): The following ensures that fields which
        # are needed for authZ policy validation are not stripped away by the
        # plugin before returning.
        original_fields, fields_to_add = self._do_field_list(
            api_common.list_args(request, 'fields'))
        filters = api_common.get_filters(request, self._attr_info,
                                         ['fields', 'sort_key', 'sort_dir',
                                          'limit', 'marker', 'page_reverse'])
        kwargs = {'filters': filters,
                  'fields': original_fields}
        sorting_helper = self._get_sorting_helper(request)
        pagination_helper = self._get_pagination_helper(request)
        sorting_helper.update_args(kwargs)
        sorting_helper.update_fields(original_fields, fields_to_add)
        pagination_helper.update_args(kwargs)
        pagination_helper.update_fields(original_fields, fields_to_add)
        if parent_id:
            kwargs[self._parent_id_name] = parent_id
        obj_getter = getattr(self._plugin, self._plugin_handlers[self.LIST])
        obj_list = obj_getter(request.context, **kwargs)
        obj_list = sorting_helper.sort(obj_list)
        obj_list = pagination_helper.paginate(obj_list)
        # Check authz
        if do_authz:
            # FIXME(salvatore-orlando): obj_getter might return references to
            # other resources. Must check authZ on them too.
            # Omit items from list that should not be visible
            obj_list = [obj for obj in obj_list
                        if policy.check(request.context,
                                        self._plugin_handlers[self.SHOW],
                                        obj,
                                        plugin=self._plugin)]
        # Use the first element in the list for discriminating which attributes
        # should be filtered out because of authZ policies
        # fields_to_add contains a list of attributes added for request policy
        # checks but that were not required by the user. They should be
        # therefore stripped
        fields_to_strip = fields_to_add or []
        if obj_list:
            fields_to_strip += self._exclude_attributes_by_policy(
                request.context, obj_list[0])
        collection = {self._collection:
                      [self._filter_attributes(
                          request.context, obj,
                          fields_to_strip=fields_to_strip)
                       for obj in obj_list]}
        pagination_links = pagination_helper.get_links(obj_list)
        if pagination_links:
            collection[self._collection + "_links"] = pagination_links
        return collection

    def _item(self, request, id, do_authz=False, field_list=None,
              parent_id=None):
        """Retrieves and formats a single element of the requested entity."""
        kwargs = {'fields': field_list}
        action = self._plugin_handlers[self.SHOW]
        if parent_id:
            kwargs[self._parent_id_name] = parent_id
        obj_getter = getattr(self._plugin, action)
        obj = obj_getter(request.context, id, **kwargs)
        # Check authz
        # FIXME(salvatore-orlando): obj_getter might return references to
        # other resources. Must check authZ on them too.
        if do_authz:
            policy.enforce(request.context, action, obj)
        return obj

    def index(self, request, **kwargs):
        """Returns a list of the requested entity."""
        parent_id = kwargs.get(self._parent_id_name)
        # Ensure policy engine is initialized
        policy.init()
        return self._items(request, True, parent_id)

    def show(self, request, id, **kwargs):
        """Returns detailed information about the requested entity."""
        try:
            # NOTE(salvatore-orlando): The following ensures that fields
            # which are needed for authZ policy validation are not stripped
            # away by the plugin before returning.
            field_list, added_fields = self._do_field_list(
                api_common.list_args(request, "fields"))
            parent_id = kwargs.get(self._parent_id_name)
            # Ensure policy engine is initialized
            policy.init()
            return {self._resource:
                    self._view(request.context,
                               self._item(request,
                                          id,
                                          do_authz=True,
                                          field_list=field_list,
                                          parent_id=parent_id),
                               fields_to_strip=added_fields)}
        except exceptions.PolicyNotAuthorized:
            # To avoid giving away information, pretend that it
            # doesn't exist
            msg = _('The resource could not be found.')
            raise webob.exc.HTTPNotFound(msg)

    def _emulate_bulk_create(self, obj_creator, request, body, parent_id=None):
        objs = []
        try:
            for item in body[self._collection]:
                kwargs = {self._resource: item}
                if parent_id:
                    kwargs[self._parent_id_name] = parent_id
                fields_to_strip = self._exclude_attributes_by_policy(
                    request.context, item)
                objs.append(self._filter_attributes(
                    request.context,
                    obj_creator(request.context, **kwargs),
                    fields_to_strip=fields_to_strip))
            return objs
        # Note(salvatore-orlando): broad catch as in theory a plugin
        # could raise any kind of exception
        except Exception:
            for obj in objs:
                obj_deleter = getattr(self._plugin,
                                      self._plugin_handlers[self.DELETE])
                try:
                    kwargs = ({self._parent_id_name: parent_id} if parent_id
                              else {})
                    obj_deleter(request.context, obj['id'], **kwargs)
                except Exception:
                    # broad catch as our only purpose is to log the exception
                    LOG.exception("Unable to undo add for %(resource)s %(id)s",
                                  {'resource': self._resource,
                                   'id': obj['id']})
            # TODO(salvatore-orlando): The object being processed when the
            # plugin raised might have been created or not in the db.
            # We need a way for ensuring that if it has been created,
            # it is then deleted
            raise

    def create(self, request, body=None, **kwargs):
        """Creates a new instance of the requested entity."""
        parent_id = kwargs.get(self._parent_id_name)
        self._notifier.info(request.context,
                            self._resource + '.create.start',
                            body)
        body = Controller.prepare_request_body(request.context, body, True,
                                               self._resource, self._attr_info,
                                               allow_bulk=self._allow_bulk)
        action = self._plugin_handlers[self.CREATE]
        # Check authz
        if self._collection in body:
            # Have to account for bulk create
            items = body[self._collection]
        else:
            items = [body]
        # Ensure policy engine is initialized
        policy.init()
        for item in items:
            policy.enforce(request.context,
                           action,
                           item[self._resource])

        def notify(create_result):
            notifier_method = self._resource + '.create.end'
            self._notifier.info(request.context,
                                notifier_method,
                                create_result)
            return create_result

        kwargs = {self._parent_id_name: parent_id} if parent_id else {}
        if self._collection in body and self._native_bulk:
            # plugin does atomic bulk create operations
            obj_creator = getattr(self._plugin, "%s_bulk" % action)
            objs = obj_creator(request.context, body, **kwargs)
            # Use first element of list to discriminate attributes which
            # should be removed because of authZ policies
            fields_to_strip = self._exclude_attributes_by_policy(
                request.context, objs[0])
            return notify({self._collection: [self._filter_attributes(
                request.context, obj, fields_to_strip=fields_to_strip)
                for obj in objs]})
        else:
            obj_creator = getattr(self._plugin, action)
            if self._collection in body:
                # Emulate atomic bulk behavior
                objs = self._emulate_bulk_create(obj_creator, request,
                                                 body, parent_id)
                return notify({self._collection: objs})
            else:
                kwargs.update({self._resource: body})
                obj = obj_creator(request.context, **kwargs)
                return notify({self._resource: self._view(request.context,
                                                          obj)})

    def delete(self, request, id, body=None, **kwargs):
        """Deletes the specified entity."""
        self._notifier.info(request.context,
                            self._resource + '.delete.start',
                            {self._resource + '_id': id})
        action = self._plugin_handlers[self.DELETE]

        # Check authz
        policy.init()
        parent_id = kwargs.get(self._parent_id_name)
        obj = self._item(request, id, parent_id=parent_id)
        try:
            policy.enforce(request.context,
                           action,
                           obj)
        except exceptions.PolicyNotAuthorized:
            # To avoid giving away information, pretend that it
            # doesn't exist
            msg = _('The resource could not be found.')
            raise webob.exc.HTTPNotFound(msg)

        obj_deleter = getattr(self._plugin, action)
        if body:
            kwargs.update({self._resource: body})
        obj_deleter(request.context, id, **kwargs)
        notifier_method = self._resource + '.delete.end'
        self._notifier.info(request.context,
                            notifier_method,
                            {self._resource + '_id': id})

    def update(self, request, id, body=None, **kwargs):
        """Updates the specified entity's attributes."""
        parent_id = kwargs.get(self._parent_id_name)
        try:
            payload = body.copy()
        except AttributeError:
            msg = _("Invalid format: %s") % request.body
            raise exceptions.BadRequest(resource='body', msg=msg)
        payload['id'] = id
        self._notifier.info(request.context,
                            self._resource + '.update.start',
                            payload)
        body = Controller.prepare_request_body(request.context, body, False,
                                               self._resource, self._attr_info,
                                               allow_bulk=self._allow_bulk)
        action = self._plugin_handlers[self.UPDATE]
        # Load object to check authz
        # but pass only attributes in the original body and required
        # by the policy engine to the policy 'brain'
        field_list = [name for (name, value) in (self._attr_info).items()
                      if (value.get('required_by_policy') or
                          value.get('primary_key') or
                          'default' not in value)]
        # Ensure policy engine is initialized
        policy.init()
        orig_obj = self._item(request, id, field_list=field_list,
                              parent_id=parent_id)
        orig_obj.update(body[self._resource])
        attribs = attributes.ATTRIBUTES_TO_UPDATE
        orig_obj[attribs] = body[self._resource].keys()
        try:
            policy.enforce(request.context,
                           action,
                           orig_obj)
        except exceptions.PolicyNotAuthorized:
            # To avoid giving away information, pretend that it
            # doesn't exist
            msg = _('The resource could not be found.')
            raise webob.exc.HTTPNotFound(msg)

        obj_updater = getattr(self._plugin, action)
        kwargs = {self._resource: body}
        if parent_id:
            kwargs[self._parent_id_name] = parent_id
        obj = obj_updater(request.context, id, **kwargs)
        result = {self._resource: self._view(request.context, obj)}
        notifier_method = self._resource + '.update.end'
        self._notifier.info(request.context, notifier_method, result)
        return result

    @staticmethod
    def _populate_tenant_id(context, res_dict, is_create):

        if (('tenant_id' in res_dict and
             res_dict['tenant_id'] != context.tenant_id and
             not context.is_admin)):
            msg = _("Specifying 'tenant_id' other than authenticated "
                    "tenant in request requires admin privileges")
            raise webob.exc.HTTPBadRequest(msg)

        if is_create and 'tenant_id' not in res_dict:
            if context.tenant_id:
                res_dict['tenant_id'] = context.tenant_id
            else:
                msg = _("Running without keystone AuthN requires "
                        "that tenant_id is specified")
                raise webob.exc.HTTPBadRequest(msg)

    @staticmethod
    def prepare_request_body(context, body, is_create, resource, attr_info,
                             allow_bulk=False):
        """Verifies required attributes are in request body.

        Also checking that an attribute is only specified if it is allowed
        for the given operation (create/update).

        Attribute with default values are considered to be optional.

        body argument must be the deserialized body.
        """
        collection = resource + "s"
        if not body:
            raise webob.exc.HTTPBadRequest(_("Resource body required"))

        LOG.debug("Request body: %(body)s",
                  {'body': strutils.mask_password(body)})
        prep_req_body = lambda x: Controller.prepare_request_body(  # noqa
            context,
            x if resource in x else {resource: x},
            is_create,
            resource,
            attr_info,
            allow_bulk)
        if collection in body:
            if not allow_bulk:
                raise webob.exc.HTTPBadRequest(_("Bulk operation "
                                                 "not supported"))
            bulk_body = [prep_req_body(item) for item in body[collection]]
            if not bulk_body:
                raise webob.exc.HTTPBadRequest(_("Resources required"))
            return {collection: bulk_body}

        res_dict = body.get(resource)
        if res_dict is None:
            msg = _("Unable to find '%s' in request body") % resource
            raise webob.exc.HTTPBadRequest(msg)

        Controller._populate_tenant_id(context, res_dict, is_create)
        Controller._verify_attributes(res_dict, attr_info)

        if is_create:  # POST
            for attr, attr_vals in (attr_info).items():
                if attr_vals['allow_post']:
                    if ('default' not in attr_vals and
                            attr not in res_dict):
                        msg = _("Failed to parse request. Required "
                                "attribute '%s' not specified") % attr
                        raise webob.exc.HTTPBadRequest(msg)
                    res_dict[attr] = res_dict.get(attr,
                                                  attr_vals.get('default'))
                else:
                    if attr in res_dict:
                        msg = _("Attribute '%s' not allowed in POST") % attr
                        raise webob.exc.HTTPBadRequest(msg)
        else:  # PUT
            for attr, attr_vals in (attr_info).items():
                if attr in res_dict and not attr_vals['allow_put']:
                    msg = _("Cannot update read-only attribute %s") % attr
                    raise webob.exc.HTTPBadRequest(msg)

        for attr, attr_vals in (attr_info).items():
            if (attr not in res_dict or
                    res_dict[attr] is attributes.ATTR_NOT_SPECIFIED):
                continue
            # Convert values if necessary
            if 'convert_to' in attr_vals:
                res_dict[attr] = attr_vals['convert_to'](res_dict[attr])
            # Check that configured values are correct
            if 'validate' not in attr_vals:
                continue
            for rule in attr_vals['validate']:
                # skip validating vnfd_id when vnfd_template is specified to
                # create vnf
                if (resource == 'vnf') and ('vnfd_template' in body['vnf'])\
                        and (attr == "vnfd_id") and is_create:
                    continue
                # skip validating vnffgd_id when vnffgd_template is provided
                if ((resource == 'vnffg')
                        and ('vnffgd_template' in body['vnffg'])
                        and (attr == 'vnffgd_id') and is_create):
                    continue
                # skip validating nsd_id when nsd_template is provided
                if (resource == 'ns') and ('nsd_template' in body['ns'])\
                        and (attr == 'nsd_id') and is_create:
                    continue
                res = attributes.validators[rule](res_dict[attr],
                                                  attr_vals['validate'][rule])
                if res:
                    msg_dict = dict(attr=attr, reason=res)
                    msg = _("Invalid input for %(attr)s. "
                            "Reason: %(reason)s.") % msg_dict
                    raise webob.exc.HTTPBadRequest(msg)
        return body

    @staticmethod
    def _verify_attributes(res_dict, attr_info):
        extra_keys = set(res_dict.keys()) - set(attr_info.keys())
        if extra_keys:
            msg = _("Unrecognized attribute(s) '%s'") % ', '.join(extra_keys)
            raise webob.exc.HTTPBadRequest(msg)


def create_resource(collection, resource, plugin, params, allow_bulk=False,
                    member_actions=None, parent=None, allow_pagination=False,
                    allow_sorting=False):
    controller = Controller(plugin, collection, resource, params, allow_bulk,
                            member_actions=member_actions, parent=parent,
                            allow_pagination=allow_pagination,
                            allow_sorting=allow_sorting)

    return wsgi_resource.Resource(controller, FAULT_MAP)



# Copyright (C) 2020 NTT DATA
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from tacker.objects import base
from tacker.objects import fields


@base.TackerObjectRegistry.register
class VimConnectionInfo(base.TackerObject, base.TackerPersistentObject):
    # Version 1.0: Initial version
    VERSION = '1.0'

    fields = {
        'id': fields.StringField(nullable=False),
        'vim_id': fields.StringField(nullable=True, default=None),
        'vim_type': fields.StringField(nullable=False),
        'interface_info': fields.DictOfNullableStringsField(nullable=True,
                                                         default={}),
        'access_info': fields.DictOfNullableStringsField(nullable=True,
                                                         default={}),
        'extra': fields.DictOfNullableStringsField(nullable=True,
                                                   default={}),
    }

    @classmethod
    def _from_dict(cls, data_dict):
        id = data_dict.get('id')
        vim_id = data_dict.get('vim_id')
        vim_type = data_dict.get('vim_type')
        access_info = data_dict.get('access_info', {})
        interface_info = data_dict.get('interface_info', {})
        extra = data_dict.get('extra', {})
        obj = cls(id=id,
                  vim_id=vim_id,
                  vim_type=vim_type,
                  interface_info=interface_info,
                  access_info=access_info,
                  extra=extra)
        return obj

    @classmethod
    def obj_from_primitive(cls, primitive, context):
        if 'tacker_object.name' in primitive:
            vim_connection_info = super(
                VimConnectionInfo, cls).obj_from_primitive(
                primitive, context)
        else:
            vim_connection_info = VimConnectionInfo._from_dict(primitive)

        return vim_connection_info

    def to_dict(self):
        return {'id': self.id,
                'vim_id': self.vim_id,
                'vim_type': self.vim_type,
                'interface_info': self.interface_info,
                'access_info': self.access_info,
                'extra': self.extra}



#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.
from oslo_utils import uuidutils


class VnfPackage:
    VNF_PACKAGE_REQ_PATH = "/vnfpkgm/v1/vnf_packages"

    @staticmethod
    def make_list_response_body():
        return [VnfPackage.make_individual_response]

    @staticmethod
    def make_individual_response_body(vnfd_id, vnf_package_hash):
        add_artifact_hash = (
            "6513f21e44aa3da349f248188a44" +
            "bc304a3653a04122d8fb4535423c8" +
            "e1d14cd6a153f735bb0982e2" +
            "161b5b5186106570c17a9" +
            "e58b64dd39390617cd5a350f78")

        sw_image_hash = (
            "6513f21e44aa3da349" +
            "f248188a44bc304a3653a04" +
            "122d8fb4535423c8e1d14c" +
            "d6a153f735bb0982e2161b5" +
            "b5186106570c17a9e58b6" +
            "4dd39390617cd5a350f78")

        data = {
            "id": uuidutils.generate_uuid(),
            "vnfdId": vnfd_id,
            "vnfProvider": "Company",
            "vnfProductName": "Sample VNF",
            "vnfSoftwareVersion": "1.0",
            "vnfdVersion": "1.0",
            "checksum": {
                "algorithm": "SHA-512",
                "hash": vnf_package_hash
            },
            "softwareImages": [
                {
                    "id": "sw_image",
                    "name": "cirros-0.5.2-x86_64-disk",
                    "provider": "Company",
                    "version": "0.5.2",
                    "checksum": {
                        "algorithm": "SHA-512",
                        "hash": sw_image_hash
                    },
                    "containerFormat": "BARE",
                    "diskFormat": "QCOW2",
                    "createdAt": "2020-09-01T12:34:56Z",
                    "minDisk": "2147483648",
                    "minRam": "268435456",
                    "size": "1073741824",
                    "userMetadata": {
                        "key": "value"
                    },
                    "imagePath": "Files/images/cirros-0.5.2-x86_64-disk.img"
                }
            ],
            "additionalArtifacts": [
                {
                    "artifactPath":
                        "Files/images/cirros-0.5.2-x86_64-disk.img",
                    "checksum": {
                        "algorithm": "SHA-512",
                        "hash": add_artifact_hash
                    },
                    "metadata": {
                        "key": "value"
                    }
                }
            ],
            "onboardingState": "ONBOARDED",
            "operationalState": "ENABLED",
            "usageState": "NOT_IN_USE",
            "userDefinedData": {
                "key": "value"
            },
            "_links": {
                "self": {
                    "href": "GetPackage URI"
                },
                "vnfd": {
                    "href": "GetVNFD URI"
                },
                "packageContent": {
                    "href": "GetPackageContent URI"
                }
            }
        }

        return data



# Copyright (C) 2020 NTT DATA
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from tacker import context
from tacker import objects
from tacker.tests.unit.db.base import SqlTestCase
from tacker.tests import uuidsentinel


class TestVnfResource(SqlTestCase):

    def setUp(self):
        super(TestVnfResource, self).setUp()
        self.context = context.get_admin_context()

    def test_obj_from_primitive_and_object_to_dict(self):
        vim_connection_dict = {'id': uuidsentinel.uuid,
                               'vim_id': uuidsentinel.uuid,
                               'vim_type': 'openstack'}
        result = objects.VimConnectionInfo.obj_from_primitive(
            vim_connection_dict, self.context)
        self.assertIsInstance(result, objects.VimConnectionInfo)
        self.assertEqual('openstack', result.vim_type)
        vim_connection_dict = result.to_dict()
        self.assertIsInstance(vim_connection_dict, dict)
        self.assertEqual('openstack', vim_connection_dict['vim_type'])



# Copyright (C) 2020 NTT DATA
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from unittest import mock

from tacker.common import exceptions
from tacker import context
from tacker import objects
from tacker.tests.unit.db.base import SqlTestCase
from tacker.tests.unit.objects import fakes
from tacker.tests import uuidsentinel


class TestVnfResource(SqlTestCase):

    def setUp(self):
        super(TestVnfResource, self).setUp()
        self.context = context.get_admin_context()
        self.vnf_instance = self._create_vnf_instance()

    def _create_vnf_instance(self):
        vnf_package = objects.VnfPackage(context=self.context,
                                         **fakes.vnf_package_data)
        vnf_package.create()

        vnf_pack_vnfd = fakes.get_vnf_package_vnfd_data(
            vnf_package.id, uuidsentinel.vnfd_id)

        vnf_pack_vnfd_obj = objects.VnfPackageVnfd(
            context=self.context, **vnf_pack_vnfd)
        vnf_pack_vnfd_obj.create()

        vnf_package.vnf_package = "ONBOARDED"
        vnf_package.save()

        vnf_instance_data = fakes.get_vnf_instance_data(
            vnf_pack_vnfd_obj.vnfd_id)
        vnf_instance = objects.VnfInstance(context=self.context,
                                           **vnf_instance_data)
        vnf_instance.create()

        return vnf_instance

    def test_create(self):
        vnf_resource = objects.VnfResource(
            context=self.context,
            **fakes.fake_vnf_resource_data(self.vnf_instance.id))
        vnf_resource.create()
        self.assertTrue(vnf_resource.id)
        self.assertTrue(vnf_resource.tenant_id)

    def test_create_with_id(self):

        vnf_resource_data = fakes.fake_vnf_resource_data(self.vnf_instance.id)
        vnf_resource_data.update({'id': uuidsentinel.uuid})

        vnf_resource_obj = objects.VnfResource(
            context=self.context, **vnf_resource_data)
        self.assertRaises(exceptions.ObjectActionError,
                          vnf_resource_obj.create)

    @mock.patch('tacker.objects.vnf_resources._vnf_resource_update')
    def test_save(self, mock_resource_update):
        vnf_resource = objects.VnfResource(
            context=self.context,
            **fakes.fake_vnf_resource_data(self.vnf_instance.id))
        vnf_resource.create()
        mock_resource_update.return_value = \
            fakes.vnf_resource_model_object(vnf_resource)
        vnf_resource.resource_name = 'fake'
        vnf_resource.save()
        mock_resource_update.assert_called_with(
            self.context, vnf_resource.id, {'resource_name': 'fake'})

    def test_save_error(self):
        vnf_resource = objects.VnfResource(
            context=self.context,
            **fakes.fake_vnf_resource_data(self.vnf_instance.id))
        vnf_resource.create()
        vnf_resource.destroy(self.context)
        vnf_resource.resource_name = 'fake'
        self.assertRaises(exceptions.VnfResourceNotFound, vnf_resource.save)

    @mock.patch('tacker.objects.vnf_resources._destroy_vnf_resource')
    def test_destroy(self, mock_vnf_destroy):
        vnf_resource = objects.VnfResource(
            context=self.context,
            **fakes.fake_vnf_resource_data(self.vnf_instance.id))
        vnf_resource.create()
        vnf_resource.destroy(self.context)
        mock_vnf_destroy.assert_called_with(self.context, vnf_resource.id)

    def test_destroy_failure_without_id(self):
        vnf_resource_obj = objects.VnfResource(context=self.context)
        self.assertRaises(exceptions.ObjectActionError,
                          vnf_resource_obj.destroy, self.context)

    def test_get_by_vnf_instance_id(self):
        vnf_resource = objects.VnfResource(
            context=self.context,
            **fakes.fake_vnf_resource_data(self.vnf_instance.id))
        vnf_resource.create()
        vnf_resource_list = objects.VnfResourceList()
        result = vnf_resource_list.get_by_vnf_instance_id(
            self.context, self.vnf_instance.id)
        self.assertIsInstance(result.objects, list)
        self.assertTrue(result.objects)



# Copyright 2018 OpenStack Foundation
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#

"""add support vnffg to ns database

Revision ID: 4747cc26b9c6
Revises: 5d490546290c
Create Date: 2018-06-27 03:18:12.227673

"""

# flake8: noqa: E402

# revision identifiers, used by Alembic.
revision = '4747cc26b9c6'
down_revision = '5d490546290c'

from alembic import op
import sqlalchemy as sa

from tacker.db import types


def upgrade(active_plugins=None, options=None):
    op.add_column('ns', sa.Column(
        'vnffg_ids', sa.TEXT(length=65535), nullable=True))
    op.add_column('vnffgs', sa.Column(
        'ns_id', types.Uuid(length=36), nullable=True))
    op.create_foreign_key('vnffg_foreign_key',
                          'vnffgs', 'ns',
                          ['ns_id'], ['id'])



# Copyright 2016 Brocade Communications System, Inc.
# All Rights Reserved.
#
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import os
import yaml

from keystoneauth1 import exceptions
from keystoneauth1 import identity
from keystoneauth1.identity import v3
from keystoneauth1 import session
from neutronclient.common import exceptions as nc_exceptions
from neutronclient.v2_0 import client as neutron_client
from oslo_config import cfg
from oslo_log import log as logging

from tacker._i18n import _
from tacker.common import log
from tacker import context as t_context
from tacker.extensions import nfvo
from tacker.keymgr import API as KEYMGR_API
from tacker.mistral import mistral_client
from tacker.nfvo.drivers.vim import abstract_vim_driver
from tacker.nfvo.drivers.vnffg import abstract_vnffg_driver
from tacker.nfvo.drivers.workflow import workflow_generator
from tacker.nfvo.nfvo_plugin import NfvoPlugin
from tacker.plugins.common import constants
from tacker.vnfm import keystone

LOG = logging.getLogger(__name__)
CONF = cfg.CONF

OPTS = [cfg.StrOpt('openstack', default='/etc/tacker/vim/fernet_keys',
                   help='Dir.path to store fernet keys.'),
        cfg.BoolOpt('use_barbican', default=False,
                    help=_('Use barbican to encrypt vim password if True, '
                           'save vim credentials in local file system '
                           'if False'))
        ]

# same params as we used in ping monitor driver
OPENSTACK_OPTS = [
    cfg.StrOpt('count', default='1',
               help=_('Number of ICMP packets to send')),
    cfg.StrOpt('timeout', default='1',
               help=_('Number of seconds to wait for a response')),
    cfg.StrOpt('interval', default='1',
               help=_('Number of seconds to wait between packets'))
]
cfg.CONF.register_opts(OPTS, 'vim_keys')
cfg.CONF.register_opts(OPENSTACK_OPTS, 'vim_monitor')

_VALID_RESOURCE_TYPES = {'network': {'client': neutron_client.Client,
                                     'cmd': 'list_networks',
                                     'vim_res_name': 'networks',
                                     'filter_attr': 'name'
                                     }
                         }

FC_MAP = {'name': 'name',
          'description': 'description',
          'eth_type': 'ethertype',
          'ip_src_prefix': 'source_ip_prefix',
          'ip_dst_prefix': 'destination_ip_prefix',
          'source_port_min': 'source_port_range_min',
          'source_port_max': 'source_port_range_max',
          'destination_port_min': 'destination_port_range_min',
          'destination_port_max': 'destination_port_range_max',
          'network_src_port_id': 'logical_source_port',
          'network_dst_port_id': 'logical_destination_port'}

CONNECTION_POINT = 'connection_points'
SFC_ENCAP = 'sfc_encap'


def config_opts():
    return [('vim_keys', OPTS), ('vim_monitor', OPENSTACK_OPTS)]


class OpenStack_Driver(abstract_vim_driver.VimAbstractDriver,
                       abstract_vnffg_driver.VnffgAbstractDriver):
    """Driver for OpenStack VIM

    OpenStack driver handles interactions with local as well as
    remote OpenStack instances. The driver invokes keystone service for VIM
    authorization and validation. The driver is also responsible for
    discovering placement attributes such as regions, availability zones
    """

    def __init__(self):
        self.keystone = keystone.Keystone()
        self.keystone.create_key_dir(CONF.vim_keys.openstack)

    def get_type(self):
        return 'openstack'

    def get_name(self):
        return 'OpenStack VIM Driver'

    def get_description(self):
        return 'OpenStack VIM Driver'

    def authenticate_vim(self, vim_obj):
        """Validate VIM auth attributes

        Initialize keystoneclient with provided authentication attributes.
        """
        verify = 'True' == vim_obj['auth_cred'].get('cert_verify', 'True')
        auth_url = vim_obj['auth_url']
        keystone_version = NfvoPlugin.validate_keystone_auth_url(
            auth_url=auth_url,
            verify=verify)
        auth_cred = self._get_auth_creds(vim_obj, keystone_version)
        return self._initialize_keystone(auth_cred)

    def _get_auth_creds(self, vim_obj, keystone_version):
        auth_cred = vim_obj['auth_cred']
        vim_project = vim_obj['vim_project']
        auth_cred['project_id'] = vim_project.get('id')
        auth_cred['project_name'] = vim_project.get('name')
        auth_cred['project_domain_name'] = vim_project.get(
            'project_domain_name')
        auth_cred['auth_url'] = vim_obj['auth_url']
        if keystone_version not in auth_cred['auth_url']:
            auth_cred['auth_url'] = auth_cred['auth_url'] + '/' + \
                keystone_version
        return auth_cred

    def _get_auth_plugin(self, **kwargs):
        auth_plugin = v3.Password(**kwargs)

        return auth_plugin

    def _initialize_keystone(self, auth):
        ks_client = self.keystone.initialize_client(**auth)
        return ks_client

    def _find_regions(self, ks_client):
        region_info = ks_client.regions.list()
        region_list = [region.id for region in region_info]
        return region_list

    def discover_placement_attr(self, vim_obj, ks_client):
        """Fetch VIM placement information

        Attributes can include regions, AZ.
        """
        try:
            regions_list = self._find_regions(ks_client)
        except (exceptions.Unauthorized, exceptions.BadRequest) as e:
            LOG.warning("Authorization failed for user")
            raise nfvo.VimUnauthorizedException(message=e.message)
        vim_obj['placement_attr'] = {'regions': regions_list}
        return vim_obj

    @log.log
    def register_vim(self, vim_obj):
        """Validate and set VIM placements."""

        if 'key_type' in vim_obj['auth_cred']:
            vim_obj['auth_cred'].pop('key_type')
        if 'secret_uuid' in vim_obj['auth_cred']:
            vim_obj['auth_cred'].pop('secret_uuid')

        ks_client = self.authenticate_vim(vim_obj)
        self.discover_placement_attr(vim_obj, ks_client)
        self.encode_vim_auth(vim_obj['id'], vim_obj['auth_cred'])
        LOG.debug('VIM registration completed for %s', vim_obj)

    @log.log
    def deregister_vim(self, vim_obj):
        """Deregister VIM from NFVO

        Delete VIM keys from file system
        """
        self.delete_vim_auth(vim_obj['id'], vim_obj['auth_cred'])

    @log.log
    def delete_vim_auth(self, vim_id, auth):
        """Delete vim information

        Delete vim key stored in file system
        """
        LOG.debug('Attempting to delete key for vim id %s', vim_id)

        if auth.get('key_type') == 'barbican_key':
            try:
                k_context = t_context.generate_tacker_service_context()
                keystone_conf = CONF.keystone_authtoken
                secret_uuid = auth['secret_uuid']
                keymgr_api = KEYMGR_API(keystone_conf.auth_url)
                keymgr_api.delete(k_context, secret_uuid)
                LOG.debug('VIM key deleted successfully for vim %s',
                          vim_id)
            except Exception as ex:
                LOG.warning('VIM key deletion failed for vim %s due to %s',
                            vim_id,
                            ex)
                raise
        else:
            key_file = os.path.join(CONF.vim_keys.openstack, vim_id)
            try:
                os.remove(key_file)
                LOG.debug('VIM key deleted successfully for vim %s',
                          vim_id)
            except OSError:
                LOG.warning('VIM key deletion failed for vim %s',
                            vim_id)

    @log.log
    def encode_vim_auth(self, vim_id, auth):
        """Encode VIM credentials

         Store VIM auth using fernet key encryption
         """
        fernet_key, fernet_obj = self.keystone.create_fernet_key()
        encoded_auth = fernet_obj.encrypt(auth['password'].encode('utf-8'))
        auth['password'] = encoded_auth

        if CONF.vim_keys.use_barbican:
            try:
                k_context = t_context.generate_tacker_service_context()
                keystone_conf = CONF.keystone_authtoken
                keymgr_api = KEYMGR_API(keystone_conf.auth_url)
                secret_uuid = keymgr_api.store(k_context, fernet_key)

                auth['key_type'] = 'barbican_key'
                auth['secret_uuid'] = secret_uuid
                LOG.debug('VIM auth successfully stored for vim %s',
                          vim_id)
            except Exception as ex:
                LOG.warning('VIM key creation failed for vim %s due to %s',
                            vim_id,
                            ex)
                raise

        else:
            auth['key_type'] = 'fernet_key'
            key_file = os.path.join(CONF.vim_keys.openstack, vim_id)
            try:
                with open(key_file, 'wb') as f:
                    f.write(fernet_key)
                    LOG.debug('VIM auth successfully stored for vim %s',
                              vim_id)
            except IOError:
                raise nfvo.VimKeyNotFoundException(vim_id=vim_id)

    @log.log
    def get_vim_resource_id(self, vim_obj, resource_type, resource_name):
        """Locates openstack resource by type/name and returns ID

        :param vim_obj: VIM info used to access openstack instance
        :param resource_type: type of resource to find
        :param resource_name: name of resource to locate
        :return: ID of resource
        """
        if resource_type in _VALID_RESOURCE_TYPES:
            res_cmd_map = _VALID_RESOURCE_TYPES[resource_type]
            client_type = res_cmd_map['client']
            cmd = res_cmd_map['cmd']
            filter_attr = res_cmd_map.get('filter_attr')
            vim_res_name = res_cmd_map['vim_res_name']
        else:
            raise nfvo.VimUnsupportedResourceTypeException(type=resource_type)

        client = self._get_client(vim_obj, client_type)
        cmd_args = {}
        if filter_attr:
            cmd_args[filter_attr] = resource_name

        try:
            resources = getattr(client, "%s" % cmd)(**cmd_args)[vim_res_name]
            LOG.debug('resources output %s', resources)
        except Exception:
            raise nfvo.VimGetResourceException(
                cmd=cmd, name=resource_name, type=resource_type)

        if len(resources) > 1:
            raise nfvo.VimGetResourceNameNotUnique(
                cmd=cmd, name=resource_name)
        elif len(resources) < 1:
            raise nfvo.VimGetResourceNotFoundException(
                cmd=cmd, name=resource_name)

        return resources[0]['id']

    @log.log
    def _get_client(self, vim_obj, client_type):
        """Initializes and returns an openstack client

        :param vim_obj: VIM Information
        :param client_type: openstack client to initialize
        :return: initialized client
        """
        verify = 'True' == vim_obj.get('cert_verify', 'True') or False
        auth_url = vim_obj['auth_url']
        keystone_version = NfvoPlugin.validate_keystone_auth_url(
            auth_url=auth_url,
            verify=verify)
        auth_cred = self._get_auth_creds(vim_obj, keystone_version)
        auth_plugin = self._get_auth_plugin(**auth_cred)
        sess = session.Session(auth=auth_plugin)
        return client_type(session=sess)

    def _translate_ip_protocol(self, ip_proto):
        if ip_proto == '1':
            return 'icmp'
        elif ip_proto == '6':
            return 'tcp'
        elif ip_proto == '17':
            return 'udp'
        else:
            return None

    def _create_classifier_params(self, fc):
        classifier_params = {}
        for field in fc:
            if field in FC_MAP:
                classifier_params[FC_MAP[field]] = fc[field]
            elif field == 'ip_proto':
                protocol = self._translate_ip_protocol(str(fc[field]))
                if not protocol:
                    raise ValueError('protocol %s not supported' % fc[field])
                classifier_params['protocol'] = protocol
            else:
                LOG.warning("flow classifier %s not supported by "
                            "networking-sfc driver", field)
        return classifier_params

    def create_flow_classifier(self, name, fc, auth_attr=None):
        if not auth_attr:
            LOG.warning("auth information required for n-sfc driver")
            return None
        fc['name'] = name
        LOG.debug('fc passed is %s', fc)

        sfc_classifier_params = self._create_classifier_params(fc)
        LOG.debug('sfc_classifier_params is %s', sfc_classifier_params)

        if len(sfc_classifier_params) > 0:
            neutronclient_ = NeutronClient(auth_attr)

            fc_id = neutronclient_.flow_classifier_create(
                sfc_classifier_params)
            return fc_id

        raise ValueError('empty match field for input flow classifier')

    def create_chain(self, name, path_id, fc_ids, vnfs, symmetrical=False,
                     correlation='mpls', auth_attr=None):
        if not auth_attr:
            LOG.warning("auth information required for n-sfc driver")
            return None

        neutronclient_ = NeutronClient(auth_attr)
        port_pairs_list = neutronclient_.port_pair_list()
        port_pair_groups_list = neutronclient_.port_pair_group_list()
        port_chains_list = neutronclient_.port_chain_list()
        port_pair_group_list = []
        new_ppgs = []
        new_pps = []

        try:
            for vnf in vnfs:
                # TODO(s3wong): once scaling is in place and VNFFG supports it
                # that model needs to be implemented to concatenate all
                # port-pairs into the port-pair-group
                # port pair group could include port-pairs from different VNFs
                if CONNECTION_POINT not in vnf:
                    LOG.warning("Chain creation failed due to missing "
                                "connection point info in VNF "
                                "%(vnfname)s", {'vnfname': vnf['name']})
                    return None
                cp_list = vnf[CONNECTION_POINT]
                num_cps = len(cp_list)
                if num_cps not in [1, 2]:
                    LOG.warning("Chain creation failed due to wrong number of "
                                "connection points: expected [1 | 2], got "
                                "%(cps)d", {'cps': num_cps})
                    return None
                if num_cps == 1:
                    ingress = cp_list[0]
                    egress = cp_list[0]
                else:
                    ingress = cp_list[0]
                    egress = cp_list[1]

                # If sfc_encap is True, pp_corr is set to correlation to
                # make use of correlation, otherwise pp_corr is set to None
                # to install SFC proxy
                sfc_encap = vnf.get(SFC_ENCAP, True)
                pp_corr = correlation if sfc_encap else None

                # valid_port_in_use function is used to find out the
                # port_pair_group_id of the existing port pair group
                # which was created by ingress and egress of current VNF
                port_pair_group_id = self.valid_port_in_use(
                    ingress, egress, port_pairs_list, port_pair_groups_list)
                if not port_pair_group_id:
                    # create the new port pair group if it is not existed
                    port_pair = dict()
                    port_pair['name'] = vnf['name'] + '-connection-points'
                    port_pair['description'] = 'port pair for ' + vnf['name']
                    port_pair['ingress'] = ingress
                    port_pair['egress'] = egress
                    port_pair['service_function_parameters'] = {
                        'correlation': pp_corr}
                    port_pair_id = neutronclient_.port_pair_create(port_pair)
                    if not port_pair_id:
                        LOG.warning("Chain creation failed due to port pair "
                                    "creation failed for vnf %(vnf)s",
                                    {'vnf': vnf['name']})
                        return None
                    new_pps.append(port_pair_id)
                    port_pair_group = {}
                    port_pair_group['name'] = vnf['name'] + '-port-pair-group'
                    port_pair_group['description'] = \
                        'port pair group for ' + vnf['name']
                    port_pair_group['port_pairs'] = []
                    port_pair_group['port_pairs'].append(port_pair_id)
                    port_pair_group_id = \
                        neutronclient_.port_pair_group_create(port_pair_group)
                    new_ppgs.append(port_pair_group_id)
                if not port_pair_group_id:
                    LOG.warning("Chain creation failed due to port pair group "
                                "creation failed for vnf "
                                "%(vnf)s", {'vnf': vnf['name']})
                    raise nfvo.CreateChainException(
                        message="Failed to create port-pair-group")
                port_pair_group_list.append(port_pair_group_id)

            # Check list port pair group between new port chain and the
            # existing port chains.  Networking-sfc does not allow to create
            # two port chains with the same port pair groups and the same order
            for pc in port_chains_list['port_chains']:
                ppg_list = pc['port_pair_groups']
                if ppg_list == port_pair_group_list:
                    # raise exception when the Vnffg path is already existing
                    raise nfvo.CreateChainException(
                        message="Vnffg path already exists")
        except nfvo.CreateChainException as e:
            # clean neutron resources such as port pair, port pair group and
            # flow classifier if we create it
            for ppg in new_ppgs:
                neutronclient_.port_pair_group_delete(ppg)
            for pp in new_pps:
                neutronclient_.port_pair_delete(pp)
            for fc_id in fc_ids:
                neutronclient_.flow_classifier_delete(fc_id)
            raise e

        # TODO(s3wong): should the chain name be given as a parameter?
        port_chain = {}
        port_chain['name'] = name + '-port-chain'
        if path_id:
            port_chain['chain_id'] = path_id
        port_chain['description'] = 'port-chain for Tacker VNFFG'
        port_chain['port_pair_groups'] = port_pair_group_list
        port_chain['flow_classifiers'] = fc_ids
        port_chain['chain_parameters'] = {}
        port_chain['chain_parameters']['symmetric'] = symmetrical
        port_chain['chain_parameters']['correlation'] = correlation
        return neutronclient_.port_chain_create(port_chain)

    def update_chain(self, chain_id, fc_ids, vnfs,
                     symmetrical=None, auth_attr=None):
        # (s3wong): chain can be updated either for
        # the list of fc and/or list of port-pair-group
        # since n-sfc driver does NOT track the ppg id
        # it will look it up (or reconstruct) from
        # networking-sfc DB --- but the caveat is that
        # the VNF name MUST be unique

        # TODO(mardim) Currently we figure out which VNF belongs to what
        # port-pair-group or port-pair through the name of VNF.
        # This is not the best approach. The best approach for the future
        # propably is to maintain in the database the ID of the
        # port-pair-group and port-pair that VNF belongs to so we can
        # implemement the update in a more robust way.

        if not auth_attr:
            LOG.warning("auth information required for n-sfc driver")
            return None

        neutronclient_ = NeutronClient(auth_attr)
        port_pairs_list = neutronclient_.port_pair_list()
        port_pair_groups_list = neutronclient_.port_pair_group_list()
        port_chains_list = neutronclient_.port_chain_list()
        new_ppgs = []
        updated_port_chain = dict()

        pc_info = neutronclient_.port_chain_show(chain_id)
        if set(fc_ids) != set(pc_info['port_chain']['flow_classifiers']):
            updated_port_chain['flow_classifiers'] = fc_ids
        old_ppgs = pc_info['port_chain']['port_pair_groups']
        old_ppgs_dict = {neutronclient_.
                    port_pair_group_show(ppg_id)['port_pair_group']['name'].
                    split('-')[0]: ppg_id for ppg_id in old_ppgs}
        past_ppgs_dict = old_ppgs_dict.copy()
        try:
            for vnf in vnfs:
                port_pair_group = {}
                port_pair = {}
                if vnf['name'] in old_ppgs_dict:
                    old_ppg_id = old_ppgs_dict.pop(vnf['name'])
                    new_ppgs.append(old_ppg_id)
                else:
                    if CONNECTION_POINT not in vnf:
                        LOG.warning("Chain update failed due to missing "
                                    "connection point info in VNF "
                                    "%(vnfname)s", {'vnfname': vnf['name']})
                        raise nfvo.UpdateChainException(
                            message="Connection point not found")
                    cp_list = vnf[CONNECTION_POINT]
                    num_cps = len(cp_list)
                    if num_cps not in [1, 2]:
                        LOG.warning("Chain update failed due to wrong number "
                                    "of connection points: expected [1 | 2],"
                                    "got %(cps)d", {'cps': num_cps})
                        raise nfvo.UpdateChainException(
                            message="Invalid number of connection points")
                    if num_cps == 1:
                        ingress = cp_list[0]
                        egress = cp_list[0]
                    else:
                        ingress = cp_list[0]
                        egress = cp_list[1]

                    # valid_port_in_use function is used to find out the
                    # port_pair_group_id of the existing port pair group
                    # which was created by ingress and egress of current VNF
                    port_pair_group_id = self.valid_port_in_use(
                        ingress, egress, port_pairs_list,
                        port_pair_groups_list)
                    if not port_pair_group_id:
                        port_pair['name'] = vnf['name'] + '-connection-points'
                        port_pair['description'] = \
                            'port pair for ' + vnf['name']
                        port_pair['ingress'] = ingress
                        port_pair['egress'] = egress
                        port_pair_id = neutronclient_.port_pair_create(
                            port_pair)
                        if not port_pair_id:
                            LOG.warning("Chain update failed due to port pair "
                                        "creation failed for "
                                        "vnf %(vnf)s", {'vnf': vnf['name']})
                            raise nfvo.UpdateChainException(
                                message="Failed to create port-pair")
                        port_pair_group['name'] = \
                            vnf['name'] + '-port-pair-group'
                        port_pair_group['description'] = \
                            'port pair group for ' + vnf['name']
                        port_pair_group['port_pairs'] = []
                        port_pair_group['port_pairs'].append(port_pair_id)
                        port_pair_group_id = neutronclient_.\
                            port_pair_group_create(port_pair_group)
                    if not port_pair_group_id:
                        LOG.warning("Chain update failed due to port pair "
                                    "group creation failed for vnf "
                                    "%(vnf)s", {'vnf': vnf['name']})
                        for pp_id in port_pair_group['port_pairs']:
                            neutronclient_.port_pair_delete(pp_id)
                        raise nfvo.UpdateChainException(
                            message="Failed to create port-pair-group")
                    new_ppgs.append(port_pair_group_id)
            for pc in port_chains_list['port_chains']:
                ppg_list = pc['port_pair_groups']
                if ppg_list == new_ppgs:
                    # raise exception when the Vnffg path already exists
                    nfvo.UpdateChainException(
                        message="Vnffg path already exists")
        except nfvo.UpdateChainException as e:
            self._delete_ppgs_and_pps(neutronclient_, new_ppgs,
                                      past_ppgs_dict, port_chains_list, fc_ids)
            raise e

        updated_port_chain['port_pair_groups'] = new_ppgs
        updated_port_chain['flow_classifiers'] = fc_ids
        try:
            pc_id = neutronclient_.port_chain_update(chain_id,
                                                     updated_port_chain)
        except (nc_exceptions.BadRequest, nfvo.UpdateChainException) as e:
            self._delete_ppgs_and_pps(neutronclient_, new_ppgs,
                                      past_ppgs_dict, port_chains_list)
            raise e
        for ppg_name in old_ppgs_dict:
            ppg_info = neutronclient_. \
                port_pair_group_show(old_ppgs_dict[ppg_name])
            ppg_inuse = self.valid_ppg_for_multiple_chain(
                ppg_info['port_pair_group']['id'], port_chains_list)
            if not ppg_inuse:
                neutronclient_.port_pair_group_delete(old_ppgs_dict[ppg_name])
                port_pairs = ppg_info['port_pair_group']['port_pairs']
                if port_pairs and len(port_pairs):
                    for j in range(0, len(port_pairs)):
                        pp_id = port_pairs[j]
                        neutronclient_.port_pair_delete(pp_id)
        return pc_id

    def delete_chain(self, chain_id, auth_attr=None):
        if not auth_attr:
            LOG.warning("auth information required for n-sfc driver")
            return None

        neutronclient_ = NeutronClient(auth_attr)
        neutronclient_.port_chain_delete(chain_id)

    def update_flow_classifier(self, chain_id, fc, auth_attr=None):
        if not auth_attr:
            LOG.warning("auth information required for n-sfc driver")
            return None

        fc_id = fc.pop('instance_id')
        fc_status = fc.pop('status')
        match_dict = fc.pop('match')
        fc.update(match_dict)

        sfc_classifier_params = self._create_classifier_params(fc)
        neutronclient_ = NeutronClient(auth_attr)
        if fc_status == constants.PENDING_UPDATE:
            fc_info = neutronclient_.flow_classifier_show(fc_id)
            for field in sfc_classifier_params:
                # If the new classifier is the same with the old one then
                # no change needed.
                if (fc_info['flow_classifier'].get(field) is not None) and \
                        (sfc_classifier_params[field] == fc_info[
                            'flow_classifier'][field]):
                    continue

                # If the new classifier has different match criteria
                # with the old one then we strip the classifier from
                # the chain we delete the old classifier and we create
                # a new one with the same name as before but with different
                # match criteria. We are not using the flow_classifier_update
                # from the n-sfc because it does not support match criteria
                # update for an existing classifier yet.
                else:
                    try:
                        self._dissociate_classifier_from_chain(chain_id,
                                                               [fc_id],
                                                               neutronclient_)
                    except Exception as e:
                        raise e
                    fc_id = neutronclient_.flow_classifier_create(
                        sfc_classifier_params)
                    if fc_id is None:
                        raise nfvo.UpdateClassifierException(
                            message="Failed to update classifiers")
                    break

        # If the new classifier is completely different from the existing
        # ones (name and match criteria) then we just create it.
        else:
            fc_id = neutronclient_.flow_classifier_create(
                sfc_classifier_params)
            if fc_id is None:
                raise nfvo.UpdateClassifierException(
                    message="Failed to update classifiers")

        return fc_id

    def _dissociate_classifier_from_chain(self, chain_id, fc_ids,
                                          neutronclient):
        pc_info = neutronclient.port_chain_show(chain_id)
        current_fc_list = pc_info['port_chain']['flow_classifiers']
        for fc_id in fc_ids:
            current_fc_list.remove(fc_id)
        pc_id = neutronclient.port_chain_update(chain_id,
                {'flow_classifiers': current_fc_list})
        if pc_id is None:
            raise nfvo.UpdateClassifierException(
                message="Failed to update classifiers")
        for fc_id in fc_ids:
            try:
                neutronclient.flow_classifier_delete(fc_id)
            except ValueError as e:
                raise e

    def remove_and_delete_flow_classifiers(self, chain_id, fc_ids,
                                           auth_attr=None):
        if not auth_attr:
            LOG.warning("auth information required for n-sfc driver")
            raise EnvironmentError('auth attribute required for'
                                   ' networking-sfc driver')
        neutronclient_ = NeutronClient(auth_attr)
        try:
            self._dissociate_classifier_from_chain(chain_id, fc_ids,
                                                   neutronclient_)
        except Exception as e:
            raise e

    def delete_flow_classifier(self, fc_id, auth_attr=None):
        if not auth_attr:
            LOG.warning("auth information required for n-sfc driver")
            raise EnvironmentError('auth attribute required for'
                                   ' networking-sfc driver')

        neutronclient_ = NeutronClient(auth_attr)
        neutronclient_.flow_classifier_delete(fc_id)

    def get_mistral_client(self, auth_dict):
        if not auth_dict:
            LOG.warning("auth dict required to instantiate mistral client")
            raise EnvironmentError('auth dict required for'
                                   ' mistral workflow driver')
        return mistral_client.MistralClient(
            keystone.Keystone().initialize_client(**auth_dict),
            auth_dict['token']).get_client()

    def prepare_and_create_workflow(self, resource, action,
                                    kwargs, auth_dict=None):
        mistral_client = self.get_mistral_client(auth_dict)
        wg = workflow_generator.WorkflowGenerator(resource, action)
        wg.task(**kwargs)
        if not wg.get_tasks():
            raise nfvo.NoTasksException(resource=resource, action=action)
        yaml.SafeDumper.ignore_aliases = lambda self, data: True
        definition_yaml = yaml.safe_dump(wg.definition)
        workflow = mistral_client.workflows.create(definition_yaml)
        return {'id': workflow[0].id, 'input': wg.get_input_dict()}

    def execute_workflow(self, workflow, auth_dict=None):
        return self.get_mistral_client(auth_dict) \
            .executions.create(
            workflow_identifier=workflow['id'],
            workflow_input=workflow['input'],
            wf_params={})

    def get_execution(self, execution_id, auth_dict=None):
        return self.get_mistral_client(auth_dict) \
            .executions.get(execution_id)

    def delete_execution(self, execution_id, auth_dict=None):
        return self.get_mistral_client(auth_dict).executions \
            .delete(execution_id, force=True)

    def delete_workflow(self, workflow_id, auth_dict=None):
        return self.get_mistral_client(auth_dict) \
            .workflows.delete(workflow_id)

    def _delete_ppgs_and_pps(self, neutronclient, new_ppgs,
                             past_ppgs_dict, pcs_list, fc_ids):
        if new_ppgs:
            for item in new_ppgs:
                if item not in past_ppgs_dict.values():
                    ppg_inuse = self.valid_ppg_for_multiple_chain(
                        item, pcs_list)
                    if not ppg_inuse:
                        # clean port pair and port pair group if
                        # it is not in used
                        new_ppg_info = neutronclient.port_pair_group_show(item)
                        neutronclient.port_pair_group_delete(item)
                        new_port_pairs = new_ppg_info['port_pair_group'][
                            'port_pairs']
                        if new_port_pairs and len(new_port_pairs):
                            for j in range(0, len(new_port_pairs)):
                                new_pp_id = new_port_pairs[j]
                                neutronclient.port_pair_delete(new_pp_id)
        # clean flow classifiers
        for fc_id in fc_ids:
            neutronclient.flow_classifier_delete(fc_id)

    def valid_port_in_use(self, ingress, egress, pps_list, ppgs_list):
        # This function checks the the ports are used or not and return the
        # port pair group id of these ports
        port_pair_list = pps_list['port_pairs']
        port_pair_group_list = ppgs_list['port_pair_groups']
        port_pair_id = None
        port_pair_group_id = None
        for pp in port_pair_list:
            if (ingress == pp['ingress']) and (egress == pp['egress']):
                port_pair_id = pp['id']
                break
        if port_pair_id:
            for ppg in port_pair_group_list:
                if port_pair_id in ppg['port_pairs']:
                    port_pair_group_id = ppg['id']
                    break
        return port_pair_group_id

    def valid_ppg_for_multiple_chain(self, ppg_id, pcs_list):
        # This function returns True if a ppg belongs to more than one
        # port chain. If not return False.
        count = 0
        for pc in pcs_list['port_chains']:
            if ppg_id in pc['port_pair_groups']:
                count = count + 1
        return True if count > 1 else False


class NeutronClient(object):
    """Neutron Client class for networking-sfc driver"""

    def __init__(self, auth_attr):
        auth_cred = auth_attr.copy()
        verify = 'True' == auth_cred.pop('cert_verify', 'True') or False
        auth = identity.Password(**auth_cred)
        sess = session.Session(auth=auth, verify=verify)
        self.client = neutron_client.Client(session=sess)

    def flow_classifier_show(self, fc_id):
        try:
            fc = self.client.show_sfc_flow_classifier(fc_id)
            if fc is None:
                raise ValueError('classifier %s not found' % fc_id)
            return fc
        except nc_exceptions.NotFound:
            LOG.error('classifier %s not found', fc_id)
            raise ValueError('classifier %s not found' % fc_id)

    def flow_classifier_create(self, fc_dict):
        LOG.debug("fc_dict passed is {fc_dict}".format(fc_dict=fc_dict))
        try:
            fc = self.client.create_sfc_flow_classifier(
                {'flow_classifier': fc_dict})
            return fc['flow_classifier']['id']
        except Exception as ex:
            LOG.error("Error while creating Flow Classifier: %s", str(ex))
            raise nfvo.FlowClassiferCreationFailed(message=str(ex))

    def flow_classifier_update(self, fc_id, update_fc):
        update_fc_dict = {'flow_classifier': update_fc}
        return self.client.update_sfc_flow_classifier(fc_id, update_fc_dict)

    def flow_classifier_delete(self, fc_id):
        try:
            self.client.delete_sfc_flow_classifier(fc_id)
        except nc_exceptions.NotFound:
            LOG.warning("fc %s not found", fc_id)
            raise ValueError('fc %s not found' % fc_id)

    def port_pair_create(self, port_pair_dict):
        try:
            pp = self.client.create_sfc_port_pair(
                {'port_pair': port_pair_dict})
        except nc_exceptions.BadRequest as e:
            LOG.error("create port pair returns %s", e)
            raise ValueError(str(e))

        if pp and len(pp):
            return pp['port_pair']['id']
        else:
            return None

    def port_pair_list(self):
        pp_list = self.client.list_sfc_port_pairs()
        return pp_list

    def port_pair_delete(self, port_pair_id):
        try:
            self.client.delete_sfc_port_pair(port_pair_id)
        except nc_exceptions.NotFound:
            LOG.warning('port pair %s not found', port_pair_id)
            raise ValueError('port pair %s not found' % port_pair_id)

    def port_pair_group_create(self, ppg_dict):
        try:
            ppg = self.client.create_sfc_port_pair_group(
                {'port_pair_group': ppg_dict})
        except nc_exceptions.BadRequest as e:
            LOG.warning('create port pair group returns %s', e)
            raise ValueError(str(e))

        if ppg and len(ppg):
            return ppg['port_pair_group']['id']
        else:
            return None

    def port_pair_group_list(self):
        ppg_list = self.client.list_sfc_port_pair_groups()
        return ppg_list

    def port_pair_group_delete(self, ppg_id):
        try:
            self.client.delete_sfc_port_pair_group(ppg_id)
        except nc_exceptions.NotFound:
            LOG.warning('port pair group %s not found', ppg_id)
            raise ValueError('port pair group %s not found' % ppg_id)

    def port_chain_create(self, port_chain_dict):
        try:
            pc = self.client.create_sfc_port_chain(
                {'port_chain': port_chain_dict})
        except nc_exceptions.BadRequest as e:
            LOG.warning('create port chain returns %s', e)
            raise ValueError(str(e))

        if pc and len(pc):
            return pc['port_chain']['id'], pc['port_chain']['chain_id']
        else:
            return None

    def port_chain_delete(self, port_chain_id):
        try:
            port_chain = self.client.show_sfc_port_chain(port_chain_id)
            if port_chain:
                self.client.delete_sfc_port_chain(port_chain_id)
                port_chain_list = \
                    self.client.list_sfc_port_chains()['port_chains']
                ppg_list = port_chain['port_chain'].get('port_pair_groups')
                if ppg_list and len(ppg_list):
                    for i in range(0, len(ppg_list)):
                        ppg_in_use = False
                        # Firstly, Tacker delete port chain, if a port pair
                        # group still belong to other port chains, Tacker
                        # will mark it as in_use and does not delete it.
                        for pc in port_chain_list:
                            if ppg_list[i] in pc['port_pair_groups']:
                                ppg_in_use = True
                                break
                        if not ppg_in_use:
                            ppg = self.client.show_sfc_port_pair_group(
                                ppg_list[i])
                            if ppg:
                                self.client.delete_sfc_port_pair_group(
                                    ppg_list[i])
                                port_pairs = \
                                    ppg['port_pair_group']['port_pairs']
                                if port_pairs and len(port_pairs):
                                    for j in range(0, len(port_pairs)):
                                        pp_id = port_pairs[j]
                                        self.client.delete_sfc_port_pair(pp_id)
        except nc_exceptions.NotFound:
            LOG.warning('port chain %s not found', port_chain_id)
            raise ValueError('port chain %s not found' % port_chain_id)

    def port_chain_update(self, port_chain_id, port_chain):
        try:
            pc = self.client.update_sfc_port_chain(port_chain_id,
                                    {'port_chain': port_chain})
        except nc_exceptions.BadRequest as e:
            LOG.warning('update port chain returns %s', e)
            raise ValueError(str(e))
        if pc and len(pc):
            return pc['port_chain']['id']
        else:
            raise nfvo.UpdateChainException(message="Failed to update "
                                                    "port-chain")

    def port_chain_list(self):
        pc_list = self.client.list_sfc_port_chains()
        return pc_list

    def port_chain_show(self, port_chain_id):
        try:
            port_chain = self.client.show_sfc_port_chain(port_chain_id)
            if port_chain is None:
                raise ValueError('port chain %s not found' % port_chain_id)

            return port_chain
        except nc_exceptions.NotFound:
            LOG.error('port chain %s not found', port_chain_id)
            raise ValueError('port chain %s not found' % port_chain_id)

    def port_pair_group_show(self, ppg_id):
        try:
            port_pair_group = self.client.show_sfc_port_pair_group(ppg_id)
            if port_pair_group is None:
                raise ValueError('port pair group %s not found' % ppg_id)

            return port_pair_group
        except nc_exceptions.NotFound:
            LOG.warning('port pair group %s not found', ppg_id)
            raise ValueError('port pair group %s not found' % ppg_id)



# All Rights Reserved.
#
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import netaddr
import requests
import time

import copy
from oslo_log import log as logging
from oslo_serialization import jsonutils
from tacker.vnfm.monitor_drivers import abstract_driver
from tacker.vnfm.monitor_drivers.zabbix import zabbix_api as zapi


LOG = logging.getLogger(__name__)


class VNFMonitorZabbix(abstract_driver.VNFMonitorAbstractDriver):
    params = ['application', 'OS']

    def __init__(self):
        self.kwargs = None
        self.vnf = None
        self.vduname = []
        self.URL = None
        self.hostinfo = {}
        self.tenant_id = None

    def get_type(self):
        """Return one of predefined type of the hosting vnf drivers."""
        plugin_type = 'zabbix'
        return plugin_type

    def get_name(self):
        """Return a symbolic name for the VNF Monitor plugin."""
        plugin_name = 'zabbix'
        return plugin_name

    def get_description(self):
        """Return description of VNF Monitor plugin."""
        plugin_descript = 'Tacker VNFMonitor Zabbix Driver'
        return plugin_descript

    def monitor_get_config(self, plugin, context, vnf):
        """Return dict of monitor configuration data.

        :param plugin:
        :param context:
        :param vnf:
        :returns: dict
        :returns: dict of monitor configuration data
        """
        return {}

    def monitor_url(self, plugin, context, vnf):
        """Return the url of vnf to monitor.

        :param plugin:
        :param context:
        :param vnf:
        :returns: string
        :returns: url of vnf to monitor
        """
        pass

    def send_post(self, query):
        response = requests.post(self.URL, headers=zapi.HEADERS,
                                 data=jsonutils.dump_as_bytes(query))
        return dict(response.json())

    @staticmethod
    def check_error(response):
        try:
            if 'result' not in response:
                raise ValueError
        except ValueError:
            LOG.error('Cannot request error : %s', response['error']['data'])

    def create_graph(self, itemid, name, nodename):
        temp_graph_api = copy.deepcopy(zapi.dGRAPH_CREATE_API)
        gitems = [{'itemid': itemid, 'color': '00AA00'}]
        temp_graph_api['auth'] = \
            self.hostinfo[nodename]['zbx_info']['zabbix_token']
        temp_graph_api['params']['gitems'] = gitems
        temp_graph_api['params']['name'] = name
        response = self.send_post(temp_graph_api)
        VNFMonitorZabbix.check_error(response)

    def create_action(self):
        for vdu in self.vduname:
            temp_action_api = copy.deepcopy(zapi.dACTION_CREATE_API)
            temp_action_api['auth'] = \
                self.hostinfo[vdu]['zbx_info']['zabbix_token']
            tempname_api = temp_action_api['params']['operations'][0]
            temp_filter = temp_action_api['params']['filter']
            for info in (self.hostinfo[vdu]['actioninfo']):
                tempname_api['opcommand_hst'][0]['hostid'] = \
                    self.hostinfo[vdu]['hostid']
                now = time.localtime()
                rtime = str(now.tm_hour) + str(now.tm_min) + str(now.tm_sec)
                temp_name = "Trigger Action " + \
                            str(
                                vdu + rtime + " " +
                                info['item'] + " " + info['action']
                            )
                temp_action_api['params']['name'] = temp_name

                if (info['action'] == 'cmd') and \
                        (info['item'] != 'os_agent_info'):

                    tempname_api['opcommand']['command'] = info['cmd-action']

                elif (info['item'] == 'os_agent_info') \
                        and (info['action'] == 'cmd'):

                    tempname_api['opcommand']['authtype'] = 0
                    tempname_api['opcommand']['username'] = \
                        self.hostinfo[vdu]['appinfo']['ssh_username']
                    tempname_api['opcommand']['password'] = \
                        self.hostinfo[vdu]['appinfo']['ssh_password']
                    tempname_api['opcommand']['type'] = 2
                    tempname_api['opcommand']['command'] = info['cmd-action']
                    tempname_api['opcommand']['port'] = 22

                    temp_filter['conditions'][0]['value'] = info['trigger_id']
                    response = self.send_post(temp_action_api)
                    VNFMonitorZabbix.check_error(response)
                    continue

                temp_filter['conditions'][0]['value'] = info['trigger_id']
                response = self.send_post(temp_action_api)
                VNFMonitorZabbix.check_error(response)

    def create_vdu_host(self):
        for vdu in self.vduname:
            temp_host_api = zapi.dHOST_CREATE_API
            temp_group_api = zapi.dGROUP_GET_API
            temp_host_api['auth'] = \
                self.hostinfo[vdu]['zbx_info']['zabbix_token']
            temp_group_api['auth'] = \
                self.hostinfo[vdu]['zbx_info']['zabbix_token']
            response = self.send_post(temp_group_api)
            gid = response['result'][0]['groupid']
            temp_host_api['params']['host'] = str(vdu)
            if type(self.hostinfo[vdu]['mgmt_ip']) is list:
                for vduip in (self.hostinfo[vdu]['mgmt_ip']):
                    temp_host_api['params']['interfaces'][0]['ip'] = vduip
                    temp_host_api['params']['templates'][0]['templateid'] = \
                        self.hostinfo[vdu]['template_id'][0]
                    temp_host_api['params']['groups'][0]['groupid'] = gid
                    response = self.send_post(temp_host_api)
            else:
                temp_host_api['params']['interfaces'][0]['ip'] = \
                    self.hostinfo[vdu]['mgmt_ip']
                temp_host_api['params']['templates'][0]['templateid'] = \
                    self.hostinfo[vdu]['template_id'][0]
                temp_host_api['params']['groups'][0]['groupid'] = gid
                response = self.send_post(temp_host_api)
            if 'error' in response:
                now = time.localtime()
                rtime = str(now.tm_hour) + str(now.tm_min) + str(now.tm_sec)
                temp_host_api['params']['host'] = str(vdu) + rtime
                response = self.send_post(temp_host_api)
            self.hostinfo[vdu]['hostid'] = response['result']['hostids'][0]

    def create_trigger(self, trigger_params, vduname):
        temp_trigger_api = copy.deepcopy(zapi.dTRIGGER_CREATE_API)
        temp_trigger_api['auth'] = \
            self.hostinfo[vduname]['zbx_info']['zabbix_token']
        temp_trigger_api['params'] = trigger_params
        temp_trigger_api['templateid'] = \
            str(
                self.hostinfo[vduname]['template_id'][0])
        response = self.send_post(temp_trigger_api)
        VNFMonitorZabbix.check_error(response)
        return response['result']

    def _create_trigger(self):

        trigger_params = []
        trig_act_pa = []
        for vdu in self.vduname:
            temp_trigger_list = copy.deepcopy(zapi.dTRIGGER_LIST)

            temp_vdu_name = self.hostinfo[vdu]['appinfo']['app_name']
            temp_vdu_port = self.hostinfo[vdu]['appinfo']['app_port']
            for para in VNFMonitorZabbix.params:
                for item in self.hostinfo[vdu]['parameters'][para]:
                    action_list = copy.deepcopy(zapi.dACTION_LIST)
                    temp_item = self.hostinfo[vdu]['parameters'][para][item]

                    if ('app_name' != item)\
                            and ('app_port' != item) \
                            and ('ssh_username' != item) \
                            and ('ssh_password' != item):

                        if 'condition' \
                                in temp_item:
                            temp_con = temp_item['condition']

                        if len(temp_con) == 2:
                            temp_comparrision = temp_con[0]
                            temp_comparrision_value = temp_con[1]
                            temp_trigger_list[item][0]['expression'] += \
                                self.hostinfo[vdu]['template_name'] + ':'\
                                + str(
                                    zapi.dITEM_KEY_COMP[item].replace(
                                        '*', str(temp_vdu_name))) \
                                + str(
                                    zapi.COMP_VALUE[temp_comparrision]) \
                                + str(
                                    temp_comparrision_value)

                        else:
                            temp_comparrision = temp_con[0]
                            if 'os_agent_info' == item:
                                temp_trigger_list[item][0]['expression'] += \
                                    self.hostinfo[vdu]['template_name'] + ':' \
                                    + str(zapi.dITEM_KEY_COMP[item])

                            else:
                                temp_trigger_list[item][0]['expression'] += \
                                    self.hostinfo[vdu]['template_name'] + ':' \
                                    + str(
                                        zapi.dITEM_KEY_COMP[item].replace(
                                            '*', str(temp_vdu_port))) \
                                    + str(
                                        zapi.COMP_VALUE[temp_comparrision])
                        if 'actionname' in temp_item:
                            trig_act_pa.append(temp_trigger_list[item][0])
                            response = self.create_trigger(trig_act_pa, vdu)
                            del trig_act_pa[:]
                            action_list['action'] = \
                                temp_item['actionname']
                            action_list['trigger_id'] = \
                                response['triggerids'][0]
                            action_list['item'] = item
                            if 'cmd' == \
                                    temp_item['actionname']:

                                action_list['cmd-action'] = \
                                    temp_item['cmd-action']
                            self.hostinfo[vdu]['actioninfo'].append(
                                action_list)

                        else:
                            trigger_params.append(
                                temp_trigger_list[item][0])

            if len(trigger_params) != 0:
                self.create_trigger(trigger_params, vdu)
                del trigger_params[:]

    def create_item(self):
        # Create _ITEM
        for vdu in self.vduname:
            temp_item_api = copy.deepcopy(zapi.dITEM_CREATE_API)
            temp_item_api['auth'] = \
                self.hostinfo[vdu]['zbx_info']['zabbix_token']
            self.hostinfo[vdu]['appinfo'] = \
                copy.deepcopy(zapi.dAPP_INFO)
            temp_app = self.hostinfo[vdu]['parameters']['application']
            temp_item_api['params']['hostid'] = \
                self.hostinfo[vdu]['template_id'][0]

            for para in VNFMonitorZabbix.params:
                if 'application' == para:
                    for app_info in temp_app:
                        self.hostinfo[vdu]['appinfo'][app_info] = \
                            temp_app[app_info]

                for item in self.hostinfo[vdu]['parameters'][para]:
                    if ('app_name' != item) and ('app_port' != item) \
                            and ('ssh_username' != item) \
                            and ('ssh_password' != item):

                        temp_item_api['params']['name'] = \
                            zapi.dITEM_KEY_INFO[item]['name']
                        temp_item_api['params']['value_type'] = \
                            zapi.dITEM_KEY_INFO[item]['value_type']

                        if item == 'app_status':
                            temp = zapi.dITEM_KEY_INFO[item]['key_']
                            temp_item_api['params']['key_'] = temp.replace(
                                '*', str(
                                    self.hostinfo[vdu]['appinfo']['app_port']))

                        elif item == 'app_memory':
                            temp = zapi.dITEM_KEY_INFO[item]['key_']
                            temp_item_api['params']['key_'] = temp.replace(
                                '*',
                                str(
                                    self.hostinfo[vdu]['appinfo']['app_name']))

                        else:
                            temp_item_api['params']['key_'] = \
                                zapi.dITEM_KEY_INFO[item]['key_']
                        response = self.send_post(temp_item_api)
                        self.create_graph(
                            response['result']['itemids'][0],
                            temp_item_api['params']['name'], vdu)
                        VNFMonitorZabbix.check_error(response)

    def create_template(self):
        temp_template_api = copy.deepcopy(zapi.dTEMPLATE_CREATE_API)

        for vdu in self.vduname:
            temp_template_api['params']['host'] = "Tacker Template " + str(vdu)
            temp_template_api['auth'] = \
                self.hostinfo[vdu]['zbx_info']['zabbix_token']
            response = self.send_post(temp_template_api)

            if 'error' in response:
                if "already exists." in response['error']['data']:
                    now = time.localtime()
                    rtime = str(now.tm_hour) + str(now.tm_min) + str(
                        now.tm_sec)
                    temp_template_api['params']['host'] = \
                        "Tacker Template " + str(vdu) + rtime
                    response = self.send_post(temp_template_api)
                    VNFMonitorZabbix.check_error(response)
            self.hostinfo[vdu]['template_id'] = \
                response['result']['templateids']
            self.hostinfo[vdu]['template_name'] =\
                temp_template_api['params']['host']

    def add_host_to_zabbix(self):

        self.create_template()
        self.create_item()
        self._create_trigger()
        self.create_vdu_host()
        self.create_action()

    def get_token_from_zbxserver(self, node):

        temp_auth_api = copy.deepcopy(zapi.dAUTH_API)
        temp_auth_api['params']['user'] = \
            self.hostinfo[node]['zbx_info']['zabbix_user']
        temp_auth_api['params']['password'] = \
            self.hostinfo[node]['zbx_info']['zabbix_pass']
        zabbixip = \
            self.hostinfo[node]['zbx_info']['zabbix_ip']
        zabbixport = \
            self.hostinfo[node]['zbx_info']['zabbix_port']
        self.URL = "http://" + zabbixip + ":" + \
                   str(zabbixport) + zapi.URL
        if netaddr.valid_ipv6(zabbixip):
            self.URL = "http://[" + zabbixip + "]:" + \
                       str(zabbixport) + zapi.URL
        response = requests.post(
            self.URL,
            headers=zapi.HEADERS,
            data=jsonutils.dump_as_bytes(temp_auth_api)
        )
        response_dict = dict(response.json())
        VNFMonitorZabbix.check_error(response_dict)
        LOG.info('Success Connect Zabbix Server')
        return response_dict['result']

    def set_zbx_info(self, node):
        self.hostinfo[node]['zbx_info'] = \
            copy.deepcopy(zapi.dZBX_INFO)
        self.hostinfo[node]['zbx_info']['zabbix_user'] = \
            self.kwargs['vdus'][node]['zabbix_username']
        self.hostinfo[node]['zbx_info']['zabbix_pass'] = \
            self.kwargs['vdus'][node]['zabbix_password']
        self.hostinfo[node]['zbx_info']['zabbix_ip'] = \
            self.kwargs['vdus'][node]['zabbix_server_ip']
        self.hostinfo[node]['zbx_info']['zabbix_port'] = \
            self.kwargs['vdus'][node]['zabbix_server_port']
        self.hostinfo[node]['zbx_info']['zabbix_token'] = \
            self.get_token_from_zbxserver(node)

    def set_vdu_info(self):
        temp_vduname = self.kwargs['vdus'].keys()
        for node in temp_vduname:
            if 'application' in \
                    self.kwargs['vdus'][node]['parameters'] \
                    and 'OS'\
                    in self.kwargs['vdus'][node]['parameters']:
                self.vduname.append(node)
                self.hostinfo[node] = copy.deepcopy(zapi.dVDU_INFO)
                self.set_zbx_info(node)
                self.hostinfo[node]['mgmt_ip'] = \
                    self.kwargs['vdus'][node]['mgmt_ip']
                self.hostinfo[node]['parameters'] = \
                    self.kwargs['vdus'][node]['parameters']
                self.hostinfo[node]['vdu_id'] = self.vnf['id']

    def add_to_appmonitor(self, vnf, kwargs):

        self.__init__()
        self.kwargs = kwargs
        self.vnf = vnf
        self.set_vdu_info()
        self.tenant_id = self.vnf['vnfd']['tenant_id']
        self.add_host_to_zabbix()

    def monitor_call(self, vnf, kwargs):
        pass

    def monitor_app_driver(self, plugin, context, vnf, service_instance):
        return self.get_name()



#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#

import netaddr
from urllib import error as urlerr
from urllib import request as urlreq

from oslo_config import cfg
from oslo_log import log as logging

from tacker._i18n import _
from tacker.common import log
from tacker.vnfm.monitor_drivers import abstract_driver


LOG = logging.getLogger(__name__)
OPTS = [
    cfg.IntOpt('retry', default=5,
               help=_('Number of times to retry')),
    cfg.IntOpt('timeout', default=1,
               help=_('Number of seconds to wait for a response')),
    cfg.IntOpt('port', default=80,
               help=_('HTTP port number to send request'))
]
cfg.CONF.register_opts(OPTS, 'monitor_http_ping')


def config_opts():
    return [('monitor_http_ping', OPTS)]


class VNFMonitorHTTPPing(abstract_driver.VNFMonitorAbstractDriver):
    def get_type(self):
        return 'http_ping'

    def get_name(self):
        return 'HTTP ping'

    def get_description(self):
        return 'Tacker HTTP Ping Driver for VNF'

    def monitor_url(self, plugin, context, vnf):
        LOG.debug('monitor_url %s', vnf)
        return vnf.get('monitor_url', '')

    def _is_pingable(self, mgmt_ip='', retry=5, timeout=5, port=80, **kwargs):
        """Checks whether the server is reachable by using urllib.

        Waits for connectivity for `timeout` seconds,
        and if connection refused, it will retry `retry`
        times.
        :param mgmt_ip: IP to check
        :param retry: times to reconnect if connection refused
        :param timeout: seconds to wait for connection
        :param port: port number to check connectivity
        :return: bool - True or False depending on pingability.
        """
        url = 'http://' + mgmt_ip + ':' + str(port)
        if netaddr.valid_ipv6(mgmt_ip):
            url = 'http://[' + mgmt_ip + ']:' + str(port)

        for retry_index in range(int(retry)):
            try:
                urlreq.urlopen(url, timeout=timeout)
                return True
            except urlerr.URLError:
                LOG.warning('Unable to reach to the url %s', url)
        return 'failure'

    @log.log
    def monitor_call(self, vnf, kwargs):
        if not kwargs['mgmt_ip']:
            return

        return self._is_pingable(**kwargs)



#    Copyright 2018 NTT DATA.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

# NOTE(bhagyashris): You may scratch your head as you see code that imports
# this module and then accesses attributes for objects such as Instance,
# etc, yet you do not see these attributes in here. Never fear, there is
# a little bit of magic. When objects are registered, an attribute is set
# on this module automatically, pointing to the newest/latest version of
# the object.


def register_all():
    # NOTE(bhagyashris): You must make sure your object gets imported in this
    # function in order for it to be registered by services that may
    # need to receive it via RPC.
    __import__('tacker.objects.cancel_mode')
    __import__('tacker.objects.heal_vnf_request')
    __import__('tacker.objects.vnf_package')
    __import__('tacker.objects.vnf_package_vnfd')
    __import__('tacker.objects.vnf_deployment_flavour')
    __import__('tacker.objects.vnf_software_image')
    __import__('tacker.objects.vnf_instance')
    __import__('tacker.objects.vnf_instantiated_info')
    __import__('tacker.objects.vim_connection')
    __import__('tacker.objects.instantiate_vnf_req')
    __import__('tacker.objects.vnf_resources')
    __import__('tacker.objects.vnf')
    __import__('tacker.objects.vnfd')
    __import__('tacker.objects.vnf_lcm_op_occs')
    __import__('tacker.objects.terminate_vnf_req')
    __import__('tacker.objects.vnf_artifact')
    __import__('tacker.objects.vnf_lcm_subscriptions')
    __import__('tacker.objects.scale_vnf_request')
    __import__('tacker.objects.grant')
    __import__('tacker.objects.grant_request')
    __import__('tacker.objects.vnfd_attribute')
    __import__('tacker.objects.change_ext_conn_req')



# Copyright (C) 2020 FUJITSU
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import os

from tacker.tests.unit import base
from tacker.tests.unit.vnfm.infra_drivers.kubernetes import fakes
from tacker.vnfm.infra_drivers.kubernetes.k8s import translate_inputs


class TestParser(base.TestCase):
    def setUp(self):
        super(TestParser, self).setUp()
        self.k8s_client_dict = fakes.fake_k8s_client_dict()
        self.vnfd_path = '../../../../etc/samples/sample_tosca_vnfc.yaml'
        self.yaml_path = os.path.join(
            os.path.dirname(os.path.abspath(__file__)),
            self.vnfd_path)
        self.vnfd_dict = {
            "tosca_definitions_version": "tosca_simple_profile_for_nfv_1_0_0",
            "description": "Demo example",
            "metadata": {
                "template_name": "sample-tosca-vnfd"},
            "topology_template": {
                "node_templates": {
                    "VDU1": {
                        "type": "tosca.nodes.nfv.VDU.Tacker",
                        "capabilities": {
                            "nfv_compute": {
                                "properties": {
                                    "num_cpus": 1,
                                    "mem_size": "512 MB",
                                    "disk_size": "1 GB"}}},
                        "properties": {
                            "vnfcs": {
                                "web_server": {
                                    "mem_size": "100 MB",
                                    "config": "config"
                                }
                            },
                            "labels": [
                                "label1:1", "label2:2"
                            ]
                        }
                    },
                    "CP1": {
                        "type": "tosca.nodes.nfv.CP.Tacker",
                        "properties": {
                            "order": 0,
                            "management": True,
                            "anti_spoofing_protection": False},
                        "requirements": [
                            {"virtualLink": {
                                "node": "VL1"}},
                            {"virtualBinding": {
                                "node": "VDU1"}}]},
                    "VL1": {
                        "type": "tosca.nodes.nfv.VL",
                        "properties": {
                            "vendor": "Tacker",
                            "network_name": "net_mgmt"}}
                }
            }
        }
        self.parser = translate_inputs.Parser(self.vnfd_dict)

    def test_loader(self):
        tosca_kube_object = self.parser.loader()
        self.assertEqual(tosca_kube_object[0].name[:8], "svc-VDU1")
        self.assertEqual(tosca_kube_object[0].containers[0].name, "web_server")
        self.assertEqual(
            tosca_kube_object[0].containers[0].mem_size,
            100000000)
        self.assertEqual(
            tosca_kube_object[0].labels, {
                'label1': '1', 'label2': '2'})
